"""Adds metrics

Revision ID: a7ad08c851d8
Revises: bbab2640b23b
Create Date: 2021-03-29 14:57:27.081092

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a7ad08c851d8"
down_revision = "bbab2640b23b"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "metrics",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("kind", sa.String(length=50), nullable=True),
        sa.Column("guild_xid", sa.BigInteger(), nullable=True),
        sa.Column("channel_xid", sa.BigInteger(), nullable=True),
        sa.Column("user_xid", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_metrics_channel_xid"), "metrics", ["channel_xid"], unique=False
    )
    op.create_index(op.f("ix_metrics_guild_xid"), "metrics", ["guild_xid"], unique=False)
    op.create_index(op.f("ix_metrics_kind"), "metrics", ["kind"], unique=False)
    op.create_index(op.f("ix_metrics_user_xid"), "metrics", ["user_xid"], unique=False)


def downgrade():
    op.drop_index(op.f("ix_metrics_user_xid"), table_name="metrics")
    op.drop_index(op.f("ix_metrics_kind"), table_name="metrics")
    op.drop_index(op.f("ix_metrics_guild_xid"), table_name="metrics")
    op.drop_index(op.f("ix_metrics_channel_xid"), table_name="metrics")
    op.drop_table("metrics")



"""Add channel motd column

Revision ID: 9172ecfb68e5
Revises: aa4e7f42cd91
Create Date: 2020-11-22 17:53:28.138208

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9172ecfb68e5"
down_revision = "aa4e7f42cd91"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("channel_settings") as b:
        b.add_column(sa.Column("cmotd", sa.String(length=255), nullable=True))


def downgrade():
    with op.batch_alter_table("channel_settings") as b:
        b.drop_column("cmotd")



"""Add games notes

Revision ID: 31d2fa4558a3
Revises: 253d7fc50dcc
Create Date: 2021-02-20 22:18:13.771616

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "31d2fa4558a3"
down_revision = "253d7fc50dcc"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("games") as b:
        b.add_column(sa.Column("note", sa.String(length=255), nullable=True))


def downgrade():
    with op.batch_alter_table("games") as b:
        b.drop_column("note")



"""Support lfg invites

Revision ID: 18b8b53f9202
Revises: 0a01d60d4c38
Create Date: 2020-07-17 11:10:19.106275

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "18b8b53f9202"
down_revision = "0a01d60d4c38"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("users") as b:
        b.add_column(
            sa.Column(
                "invite_confirmed",
                sa.Boolean(),
                server_default=sa.false(),
                nullable=False,
            ),
        )
        b.add_column(
            sa.Column("invited", sa.Boolean(), server_default=sa.false(), nullable=False),
        )


def downgrade():
    with op.batch_alter_table("users") as b:
        b.drop_column("invited")
        b.drop_column("invite_confirmed")



"""Removes server voice category id tracking

Revision ID: b2e707d342be
Revises: 71ae365b621c
Create Date: 2020-10-16 21:36:47.067762

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b2e707d342be"
down_revision = "71ae365b621c"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("voice_category_xid")


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(sa.Column("voice_category_xid", sa.BIGINT(), nullable=True))



"""Adds setting for allow_default_avatars

Revision ID: b9390c12b27b
Revises: b2e707d342be
Create Date: 2020-10-18 14:19:03.532665

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b9390c12b27b"
down_revision = "b2e707d342be"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(
            sa.Column(
                "allow_default_avatars",
                sa.Boolean(),
                server_default=sa.true(),
                nullable=False,
            ),
        )


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("allow_default_avatars")



import logging
from typing import Any, Optional, Union, cast

import discord

from spellbot.assets import s
from spellbot.constants import EMOJI_FAIL, EMOJI_OK

logger = logging.getLogger(__name__)

ChannelType = Union[
    discord.CategoryChannel,
    discord.DMChannel,
    discord.GroupChannel,
    discord.StoreChannel,
    discord.TextChannel,
    discord.VoiceChannel,
]

MentionableChannelType = Union[
    discord.CategoryChannel,
    discord.GroupChannel,
    discord.StoreChannel,
    discord.TextChannel,
    discord.VoiceChannel,
]


def _user_or_guild_log_part(message: discord.Message) -> str:  # pragma: no cover
    if hasattr(message, "guild"):
        return f"guild {cast(Any, message.guild).id}"
    return "DM"


async def safe_remove_reaction(
    message: discord.Message, emoji: str, user: discord.User
) -> None:
    try:
        await message.remove_reaction(emoji, user)
    except discord.errors.Forbidden:
        await message.channel.send(s("reaction_permissions_required"))
    except (
        discord.errors.DiscordServerError,
        discord.errors.HTTPException,
        discord.errors.InvalidArgument,
        discord.errors.NotFound,
    ) as e:
        logger.exception(
            "warning: discord (%s): could not remove reaction: %s",
            _user_or_guild_log_part(message),
            e,
        )


async def safe_clear_reactions(message: discord.Message) -> None:
    try:
        await message.clear_reactions()
    except discord.errors.Forbidden:
        await message.channel.send(s("reaction_permissions_required"))
    except (
        discord.errors.DiscordServerError,
        discord.errors.HTTPException,
    ) as e:
        logger.exception(
            "warning: discord (%s): could not clear reactions: %s",
            _user_or_guild_log_part(message),
            e,
        )


async def safe_react_emoji(message: discord.Message, emoji: str) -> None:
    try:
        await message.add_reaction(emoji)
    except discord.errors.Forbidden:
        await message.channel.send(s("reaction_permissions_required"))
    except discord.errors.NotFound as e:
        # This isn't really an issue we need to warn about in the logs.
        logger.debug(
            "debug: discord (%s): could not react to message: %s",
            _user_or_guild_log_part(message),
            e,
        )
    except (
        discord.errors.DiscordServerError,
        discord.errors.HTTPException,
        discord.errors.InvalidArgument,
    ) as e:
        logger.exception(
            "warning: discord (%s): could not react to message: %s",
            _user_or_guild_log_part(message),
            e,
        )


async def safe_react_ok(message: discord.Message) -> None:
    await safe_react_emoji(message, EMOJI_OK)


async def safe_react_error(message: discord.Message) -> None:
    await safe_react_emoji(message, EMOJI_FAIL)


async def safe_fetch_message(
    channel: ChannelType, message_xid: int, guild_xid: int
) -> Optional[discord.Message]:
    if isinstance(
        channel, (discord.VoiceChannel, discord.CategoryChannel, discord.StoreChannel)
    ):
        return None

    try:
        return await channel.fetch_message(message_xid)
    except discord.errors.NotFound as e:
        # An admin could have deleted the message. This bot recovers from
        # this situation by simply creating a new game post. This happens
        # often enough that having it be a warning log is too verbose.
        logger.debug(
            "debug: discord (guild %s): could not fetch message: %s",
            guild_xid,
            e,
        )
    except (
        discord.errors.DiscordServerError,
        discord.errors.Forbidden,
        discord.errors.HTTPException,
    ) as e:
        logger.exception(
            "warning: discord (guild %s): could not fetch message: %s",
            guild_xid,
            e,
        )
    return None


async def safe_fetch_channel(
    client: discord.Client, channel_xid: int, guild_xid: int
) -> Optional[ChannelType]:
    channel = client.get_channel(channel_xid)
    if channel:
        return channel

    try:
        return await client.fetch_channel(channel_xid)
    except discord.errors.NotFound as e:
        # This isn't really an issue we need to warn about in the logs.
        logger.debug(
            "debug: discord (guild %s): could not fetch channel: %s",
            guild_xid,
            e,
        )
    except (
        discord.errors.DiscordServerError,
        discord.errors.Forbidden,
        discord.errors.HTTPException,
        discord.errors.InvalidData,
        discord.errors.NotFound,
    ) as e:
        logger.exception(
            "warning: discord (guild %s): could not fetch channel: %s",
            guild_xid,
            e,
        )
    return None


async def safe_fetch_user(
    client: discord.Client, user_xid: int
) -> Optional[discord.User]:
    user = client.get_user(user_xid)
    if user:
        return user

    try:
        return await client.fetch_user(user_xid)
    except (discord.errors.NotFound, discord.errors.HTTPException) as e:
        logger.exception("warning: discord: could not fetch user: %s", e)
    return None


async def safe_edit_message(
    message: discord.Message, *, reason: str = None, **options
) -> None:
    try:
        await message.edit(reason=reason, **options)
    except (
        discord.errors.DiscordServerError,
        discord.errors.Forbidden,
        discord.errors.HTTPException,
        discord.errors.InvalidArgument,
    ) as e:
        logger.exception(
            "warning: discord (%s): could not edit message: %s",
            _user_or_guild_log_part(message),
            e,
        )


async def safe_delete_message(message: discord.Message) -> None:
    try:
        await message.delete()
    except (
        discord.errors.DiscordServerError,
        discord.errors.Forbidden,
        discord.errors.HTTPException,
        discord.errors.NotFound,
    ) as e:
        logger.exception(
            "warning: discord (%s): could not delete message: %s",
            _user_or_guild_log_part(message),
            e,
        )


async def safe_send_user(user: discord.User, *args, **kwargs) -> None:
    try:
        if not hasattr(user, "send"):
            # Very rarely we get a ClientUser object here instead of a User. I have
            # no idea why this happens but a ClientUser does not have a send() method.
            # For now let's log this and return nothing.
            logger.warning(
                "warning: discord (DM): could not send message to ClientUser (%s)",
                str(user),
            )
            return None
        await user.send(*args, **kwargs)
    except discord.errors.Forbidden as e:
        # User may have the bot blocked or they may have DMs only allowed for friends.
        # Generally speaking, we can safely ignore this sort of error.
        logger.debug(
            "debug: discord (DM): could not send message to user (%s): %s",
            str(user),
            e,
        )
    except (
        discord.errors.DiscordServerError,
        discord.errors.HTTPException,
        discord.errors.InvalidArgument,
    ) as e:
        logger.exception(
            "warning: discord (DM): could not send message to user (%s): %s",
            str(user),
            e,
        )


async def safe_create_voice_channel(
    client: discord.Client,
    guild_xid: int,
    name: str,
    category: Optional[discord.CategoryChannel] = None,
) -> Optional[discord.VoiceChannel]:
    guild = client.get_guild(guild_xid)
    if not guild:
        return None

    try:
        return await guild.create_voice_channel(name, category=category)
    except (
        discord.errors.DiscordServerError,
        discord.errors.Forbidden,
        discord.errors.HTTPException,
        discord.errors.InvalidArgument,
    ) as e:
        logger.exception(
            "warning: discord (guild %s): could not create voice channel: %s",
            guild_xid,
            e,
        )
    return None


async def safe_create_category_channel(
    client: discord.Client, guild_xid: int, name: str
) -> Optional[discord.CategoryChannel]:
    guild = client.get_guild(guild_xid)
    if not guild:
        return None

    try:
        return await guild.create_category_channel(name)
    except (
        discord.errors.DiscordServerError,
        discord.errors.Forbidden,
        discord.errors.HTTPException,
        discord.errors.InvalidArgument,
    ) as e:
        logger.exception(
            "warning: discord (guild %s): could not create category channel: %s",
            guild_xid,
            e,
        )
    return None


async def safe_delete_channel(channel: ChannelType, guild_xid: int) -> bool:
    try:
        await channel.delete()  # type: ignore
        return True
    except (
        discord.errors.DiscordServerError,
        discord.errors.Forbidden,
        discord.errors.HTTPException,
        discord.errors.NotFound,
    ) as e:
        logger.exception(
            "warning: discord (guild %s): could not delete channel: %s",
            guild_xid,
            e,
        )
    return False


async def safe_create_invite(
    channel: discord.VoiceChannel, guild_xid: int, max_age: int = 0
) -> Optional[str]:
    try:
        invite = await channel.create_invite(max_age=max_age)
        return cast(str, invite.url)
    except (
        discord.errors.DiscordServerError,
        discord.errors.HTTPException,
    ) as e:
        logger.exception(
            "warning: discord (guild %s): could create channel invite: %s",
            guild_xid,
            e,
        )
    return None


async def safe_fetch_guild(
    client: discord.Client, guild_xid: int
) -> Optional[discord.Guild]:
    guild = client.get_guild(guild_xid)
    if guild:
        return guild

    try:
        return await client.fetch_guild(guild_xid)
    except (
        discord.errors.DiscordServerError,
        discord.errors.Forbidden,
        discord.errors.HTTPException,
    ) as e:
        logger.exception(
            "warning: discord (guild %s): could not fetch guild: %s",
            guild_xid,
            e,
        )
    return None


async def safe_send_channel(
    messageable: Union[discord.TextChannel, discord.DMChannel],
    content: Optional[str] = None,
    *,
    embed: Optional[discord.Embed] = None,
    file: Optional[discord.File] = None,
) -> Optional[discord.Message]:
    try:
        return await messageable.send(content, embed=embed, file=file)
    except (
        discord.errors.DiscordServerError,
        discord.errors.Forbidden,
        discord.errors.HTTPException,
        discord.errors.InvalidArgument,
    ) as e:
        logger.exception(
            "warning: discord: could not send message to channel: %s",
            e,
        )
    return None



from os.path import dirname, realpath
from pathlib import Path
from string import Template
from typing import Optional, cast

from yaml import load

try:
    from yaml import CLoader as Loader
except ImportError:  # pragma: no cover
    from yaml import Loader  # type: ignore

PACKAGE_ROOT = Path(dirname(realpath(__file__)))
ASSETS_DIR = PACKAGE_ROOT / "assets"
STRINGS_DATA_FILE = ASSETS_DIR / "strings.yaml"
ASSET_FILES = [STRINGS_DATA_FILE]


def load_strings() -> dict:
    with open(STRINGS_DATA_FILE) as f:
        return cast(dict, load(f, Loader=Loader))


__strings: Optional[dict] = None


def s(key: str, **kwargs) -> str:
    """Returns a string from strings.yaml asset with subsitutions."""
    global __strings
    if not __strings:
        __strings = load_strings()

    data = __strings.get(key, "")
    assert data, f"error: missing strings key: {key}"
    return Template(data).substitute(kwargs)



"""Track message external id for games

Revision ID: 6a1ea19d138f
Revises: 1013ebad5050
Create Date: 2020-07-14 14:58:48.950242

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6a1ea19d138f"
down_revision = "1013ebad5050"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("games") as b:
        b.add_column(sa.Column("message_xid", sa.BigInteger(), nullable=True))


def downgrade():
    with op.batch_alter_table("games") as b:
        b.drop_column("message_xid")



from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy.sql.sqltypes import DateTime, NullType, String

text = str
int_type = int
str_type = str


class StringLiteral(String):
    """Teach SA how to literalize various things."""

    def literal_processor(self, dialect):
        super_processor = super(StringLiteral, self).literal_processor(dialect)

        def process(value):
            if isinstance(value, int_type):
                return text(value)
            if not isinstance(value, str_type):
                value = text(value)
            result = super_processor(value)
            return result

        return process


class LiteralDialect(DefaultDialect):
    colspecs = {
        # prevent various encoding explosions
        String: StringLiteral,
        # teach SA about how to literalize a datetime
        DateTime: StringLiteral,
        # don't format py2 long integers to NULL
        NullType: StringLiteral,
    }


def literalquery(statement):
    """NOTE: This is entirely insecure. DO NOT execute the resulting strings."""
    import sqlalchemy.orm

    if isinstance(statement, sqlalchemy.orm.Query):
        statement = statement.statement
    return statement.compile(
        dialect=LiteralDialect(),
        compile_kwargs={"literal_binds": True},
    ).string



"""Adds teams

Revision ID: 95d86da7b3c7
Revises: b887e657fb80
Create Date: 2020-08-19 12:41:00.320532

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "95d86da7b3c7"
down_revision = "b887e657fb80"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "teams",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("guild_xid", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.String(length=50), nullable=False),
        sa.ForeignKeyConstraint(["guild_xid"], ["servers.guild_xid"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_teams_guild_xid"), "teams", ["guild_xid"], unique=False)


def downgrade():
    op.drop_index(op.f("ix_teams_guild_xid"), table_name="teams")
    op.drop_table("teams")



"""Adds links setting

Revision ID: 7a2698d54c8a
Revises: 836579a7e45f
Create Date: 2020-08-12 18:56:32.420773

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "7a2698d54c8a"
down_revision = "836579a7e45f"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(
            sa.Column(
                "links",
                sa.String(length=10),
                server_default=sa.text("'public'"),
                nullable=False,
            )
        )


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("links")



from __future__ import annotations

import asyncio
import logging
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import redis

if TYPE_CHECKING:  # pragma: no cover
    from spellbot import SpellBot

from spellbot.data import Game, Server, User
from spellbot.operations import safe_delete_channel, safe_fetch_channel

logger = logging.getLogger(__name__)


async def cleanup_expired_games(bot: SpellBot) -> None:
    """Culls games older than the given window of minutes."""
    logger.info("starting expired games cleanup task...")

    to_delete_args = []
    async with bot.session() as session:
        expired = Game.expired(session)
        for game in expired:
            if not game.is_expired():
                continue

            if game.guild_xid and game.channel_xid and game.message_xid:
                to_delete_args.append(
                    {
                        "guild_xid": game.guild_xid,
                        "channel_xid": game.channel_xid,
                        "message_xid": game.message_xid,
                    }
                )

            for user in game.users:
                # Make sure the user is still waiting and still in the
                # game that's being deleted, they could be in a new
                # game now due to how async processing works.
                if user.waiting and user.game_id == game.id:
                    user.game_id = None

            # cascade delete tag associations
            game.tags = []  # type: ignore
            session.delete(game)
        session.commit()

    for args in to_delete_args:
        async with bot.channel_lock(args["channel_xid"]):
            await bot.try_to_delete_message(**args)


async def cleanup_old_voice_channels(bot: SpellBot) -> None:
    """Checks for and deletes any bot created voice channels that are empty."""
    logger.info("starting old voice channels cleanup task...")
    async with bot.session() as session:
        for game in Game.voiced(session):
            assert game.voice_channel_xid
            chan = await safe_fetch_channel(bot, game.voice_channel_xid, game.guild_xid)
            if not chan:
                game.voice_channel_xid = None  # type: ignore
                game.voice_channel_invite = None  # type: ignore
                continue

            empty_or_really_old = (
                not chan.voice_states.keys()  # type: ignore
                or datetime.utcnow() >= game.updated_at + timedelta(hours=7)
            )
            if empty_or_really_old:
                success = await safe_delete_channel(chan, game.guild_xid)
                if success:
                    game.voice_channel_xid = None  # type: ignore
                    game.voice_channel_invite = None  # type: ignore
                else:
                    guild = bot.get_guild(game.guild_xid)
                    perms = guild.me.permissions_in(chan)
                    chan_name = chan.name  # type: ignore
                    msg = (
                        "bot permissions in channel"
                        f" ({chan_name}:{game.voice_channel_xid}):"
                        f" {perms.value}"
                    )
                    logger.info(msg)
        session.commit()


async def cleanup_started_games(bot: SpellBot) -> None:
    """Culls games older than the given window of minutes."""
    logger.info("starting started games cleanup task...")
    async with bot.session() as session:
        games = session.query(Game).filter(Game.status == "started").all()
        for game in games:
            game.tags = []  # cascade delete tag associations
            session.delete(game)
        session.commit()


async def update_metrics(bot: SpellBot) -> None:
    if not bot.metrics_db:
        return

    logger.info("starting update metrics task...")
    async with bot.session() as session:
        active = len(bot.guilds)
        games = session.query(Game).filter(Game.status == "started").count()
        servers = session.query(Server).count()
        users = session.query(User).count()
        recent_games = Game.recent_metrics(session)
        recent_users = User.recent_metrics(session)
        recent_servers = Server.recent_metrics(session)
        recent_metrics = {
            **recent_games,
            **recent_users,
            **recent_servers,
        }
        try:
            bot.metrics_db.mset(
                {
                    "servers": servers,
                    "games": games,
                    "users": users,
                    "active": active,
                }
            )
            if recent_metrics:
                bot.metrics_db.mset(recent_metrics)
        except redis.exceptions.RedisError as e:
            logger.exception("redis error: %s", e)


async def update_average_wait_times(bot: SpellBot) -> None:
    logger.info("starting update average wait times task...")
    async with bot.session() as session:
        bot.average_wait_times = {}
        avgs = Game.average_wait_times(session)
        for avg in avgs:
            bot.average_wait_times[f"{avg[0]}-{avg[1]}"] = float(avg[2])


BACKROUND_TASK_SPECS = [
    {"interval": 120, "function": cleanup_expired_games},  # 2 minutes
    {"interval": 1800, "function": update_average_wait_times},  # 30 minutes
    {"interval": 3600, "function": update_metrics},  # 1 hour
    {"interval": 600, "function": cleanup_old_voice_channels},  # 10 minutes
    # Make cleanup_started_games manually triggered?
    # {"interval": 14400, "function": "cleanup_started_games"}, # 4 hours
]


def begin_background_tasks(bot: SpellBot) -> None:
    """Start up any periodic background tasks."""
    for task_spec in BACKROUND_TASK_SPECS:

        def task_context(spec: dict) -> None:
            INTERVAL = spec["interval"]

            async def task_runner() -> None:
                while True:
                    await spec["function"](bot)
                    await asyncio.sleep(INTERVAL)

            bot.loop.create_task(task_runner())

        task_context(task_spec)



"""Adds a bunch of indexes

Revision ID: 836579a7e45f
Revises: d42e90e48b68
Create Date: 2020-07-25 17:46:12.616501

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "836579a7e45f"
down_revision = "d42e90e48b68"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "new_games_tags",
        sa.Column("game_id", sa.Integer(), nullable=True),
        sa.Column("tag_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["game_id"], ["games.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["tag_id"], ["tags.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("game_id", "tag_id", name="game_tag_pk"),
    )
    op.get_bind().execute("INSERT INTO new_games_tags SELECT * FROM games_tags;")
    op.drop_table("games_tags")
    op.rename_table("new_games_tags", "games_tags")

    op.create_index(
        op.f("ix_channels_guild_xid"), "channels", ["guild_xid"], unique=False
    )
    op.create_index(op.f("ix_games_channel_xid"), "games", ["channel_xid"], unique=False)
    op.create_index(op.f("ix_games_event_id"), "games", ["event_id"], unique=False)
    op.create_index(op.f("ix_games_guild_xid"), "games", ["guild_xid"], unique=False)
    op.create_index(op.f("ix_games_size"), "games", ["size"], unique=False)
    op.create_index(op.f("ix_games_status"), "games", ["status"], unique=False)
    op.create_index(op.f("ix_games_system"), "games", ["system"], unique=False)
    op.create_index(op.f("ix_games_updated_at"), "games", ["updated_at"], unique=False)
    op.create_index(op.f("ix_users_game_id"), "users", ["game_id"], unique=False)


def downgrade():
    op.drop_index(op.f("ix_users_game_id"), table_name="users")
    op.drop_index(op.f("ix_games_updated_at"), table_name="games")
    op.drop_index(op.f("ix_games_system"), table_name="games")
    op.drop_index(op.f("ix_games_status"), table_name="games")
    op.drop_index(op.f("ix_games_size"), table_name="games")
    op.drop_index(op.f("ix_games_guild_xid"), table_name="games")
    op.drop_index(op.f("ix_games_event_id"), table_name="games")
    op.drop_index(op.f("ix_games_channel_xid"), table_name="games")
    op.drop_index(op.f("ix_channels_guild_xid"), table_name="channels")

    op.create_table(
        "old_games_tags",
        sa.Column("game_id", sa.Integer(), nullable=True),
        sa.Column("tag_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["game_id"], ["games.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["tag_id"], ["tags.id"], ondelete="CASCADE"),
    )
    op.get_bind().execute("INSERT INTO old_games_tags SELECT * FROM games_tags;")
    op.drop_table("games_tags")
    op.rename_table("old_games_tags", "games_tags")



"""Adds server option for tags enabled

Revision ID: 9798bbb9c325
Revises: 8cb01e86be05
Create Date: 2020-10-14 08:58:51.745340

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9798bbb9c325"
down_revision = "8cb01e86be05"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(
            sa.Column(
                "tags_enabled", sa.Boolean(), server_default=sa.true(), nullable=False
            ),
        )


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("tags_enabled")



"""Add cached server name

Revision ID: 6c1db37e135b
Revises: ad589a3d7585
Create Date: 2020-10-27 20:33:01.024610

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6c1db37e135b"
down_revision = "ad589a3d7585"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(sa.Column("cached_name", sa.String(length=50), nullable=True))


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("cached_name")



"""Adds created_at to reports

Revision ID: b887e657fb80
Revises: a86e8442192a
Create Date: 2020-08-16 17:59:04.926717

"""
from datetime import datetime

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b887e657fb80"
down_revision = "a86e8442192a"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_index(op.f("ix_reports_game_id"), table_name="reports")
    op.drop_table("reports")

    op.create_table(
        "reports",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("game_id", sa.Integer(), nullable=False),
        sa.Column("report", sa.String(length=255), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False, default=datetime.utcnow),
        sa.ForeignKeyConstraint(["game_id"], ["games.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_reports_game_id"), "reports", ["game_id"], unique=False)


def downgrade():
    with op.batch_alter_table("reports") as b:
        b.drop_column("created_at")



"""Adds created_at for servers and users

Revision ID: 71ae365b621c
Revises: 9798bbb9c325
Create Date: 2020-10-14 10:50:20.748741

"""
from datetime import datetime

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "71ae365b621c"
down_revision = "9798bbb9c325"
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()

    # existing servers will have a created_at set to now
    with op.batch_alter_table("servers") as b:
        b.add_column(sa.Column("created_at", sa.DateTime(), nullable=True))
    conn.execute(sa.text("UPDATE servers SET created_at = :now"), now=datetime.utcnow())
    with op.batch_alter_table("servers") as b:
        b.alter_column("created_at", nullable=False)

    # existing users will have a created_at set to now
    with op.batch_alter_table("users") as b:
        b.add_column(sa.Column("created_at", sa.DateTime(), nullable=True))
    conn.execute(sa.text("UPDATE users SET created_at = :now"), now=datetime.utcnow())
    with op.batch_alter_table("users") as b:
        b.alter_column("created_at", nullable=False)


def downgrade():
    with op.batch_alter_table("users") as b:
        b.drop_column("created_at")

    with op.batch_alter_table("servers") as b:
        b.drop_column("created_at")



"""Allow tags for specific channels

Revision ID: 42ab4d026574
Revises: d7a58c7aaeec
Create Date: 2021-02-19 10:34:07.039733

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "42ab4d026574"
down_revision = "d7a58c7aaeec"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("channel_settings") as b:
        b.add_column(sa.Column("tags_enabled", sa.Boolean(), nullable=True))


def downgrade():
    with op.batch_alter_table("channel_settings") as b:
        b.drop_column("tags_enabled")



from __future__ import annotations

import json
import logging
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, cast

import alembic  # type: ignore
import alembic.command  # type: ignore
import alembic.config  # type: ignore
import discord
import humanize  # type: ignore
from sqlalchemy import (
    BigInteger,
    Boolean,
    Column,
    DateTime,
    Float,
    ForeignKey,
    Integer,
    String,
    Table,
    and_,
    between,
)
from sqlalchemy import cast as sql_cast
from sqlalchemy import create_engine, false, func, or_, text, true
from sqlalchemy.engine.base import Connection
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session, relationship, sessionmaker
from sqlalchemy.sql.expression import asc, desc, distinct
from sqlalchemy.sql.schema import UniqueConstraint
from sqlalchemy.sql.sqltypes import Numeric

from spellbot.constants import (
    EMOJI_DROP_GAME,
    EMOJI_JOIN_GAME,
    THUMB_URL,
    VOICE_INVITE_EXPIRE_TIME_S,
)

PACKAGE_ROOT = Path(__file__).resolve().parent
ASSETS_DIR = PACKAGE_ROOT / "assets"
ALEMBIC_INI = ASSETS_DIR / "alembic.ini"
VERSIONS_DIR = PACKAGE_ROOT / "versions"

logger = logging.getLogger(__name__)
Base = declarative_base()


class Server(Base):
    __tablename__ = "servers"
    guild_xid = Column(BigInteger, primary_key=True, nullable=False)
    cached_name = Column(String(50))
    created_at = Column(DateTime, nullable=False, default=datetime.utcnow)
    prefix = Column(String(10), nullable=False, default="!")
    expire = Column(Integer, nullable=False, server_default=text("30"))  # minutes
    links = Column(String(10), nullable=False, server_default=text("'public'"))
    show_spectate_link = Column(Boolean, nullable=False, server_default=false())
    motd = Column(String(10), nullable=False, server_default=text("'both'"))
    power_enabled = Column(Boolean, nullable=False, server_default=true())
    tags_enabled = Column(Boolean, nullable=False, server_default=true())
    queue_time_enabled = Column(Boolean, nullable=False, server_default=true())
    create_voice = Column(Boolean, nullable=False, server_default=false())
    smotd = Column(String(255))
    voice_category_prefix = Column(String(40))
    games = relationship("Game", back_populates="server", uselist=True)
    channels = relationship("Channel", back_populates="server", uselist=True)
    auto_verify_channels = relationship(
        "AutoVerifyChannel", back_populates="server", uselist=True
    )
    unverified_only_channels = relationship(
        "UnverifiedOnlyChannel", back_populates="server", uselist=True
    )
    channel_settings = relationship(
        "ChannelSettings", back_populates="server", uselist=True, lazy="dynamic"
    )
    teams = relationship("Team", back_populates="server", uselist=True)

    def bot_allowed_in(self, channel_xid: int) -> bool:
        return not self.channels or any(
            channel.channel_xid == channel_xid for channel in self.channels
        )

    def channel_settings_for(self, channel_xid: int) -> Optional[ChannelSettings]:
        filters = {"channel_xid": channel_xid}
        return self.channel_settings.filter_by(**filters).one_or_none()  # type: ignore

    @classmethod
    def recent_metrics(cls, session: Session) -> dict:
        data = [
            row[1]
            for row in (
                session.query(
                    func.date(Server.created_at).label("day"),
                    func.count(Server.guild_xid),
                )
                .filter(
                    Server.created_at >= datetime.utcnow() - timedelta(days=5),
                )
                .group_by("day")
                .order_by(desc("day"))
                .all()
            )
        ]
        return {f"servers_{i}": count for i, count in enumerate(data)}

    def games_data(self) -> Dict[str, List[str]]:
        data: Dict[str, List[str]] = {
            "id": [],
            "size": [],
            "status": [],
            "message": [],
            "system": [],
            "channel_xid": [],
            "url": [],
            "event_id": [],
            "created_at": [],
            "tags": [],
        }
        for game in sorted(self.games, key=lambda game: game.id):
            if game.status == "pending":
                continue
            tags_str = f"{','.join(sorted(tag.name for tag in game.tags))}"
            tags_str = (
                f'"{tags_str}"' if len(cast(List[Tag], game.tags)) > 1 else tags_str
            )
            event_id = game.event.id if game.event else None
            data["id"].append(str(game.id))
            data["size"].append(str(game.size))
            data["status"].append(game.status)
            data["message"].append(game.message or "")
            data["system"].append(game.system)
            data["channel_xid"].append(str(game.channel_xid) if game.channel_xid else "")
            data["url"].append(game.url or "")
            data["event_id"].append(str(event_id) if event_id else "")
            data["created_at"].append(str(game.created_at))
            data["tags"].append(tags_str)
        return data

    def __repr__(self) -> str:
        return json.dumps(
            {
                "guild_xid": self.guild_xid,
                "prefix": self.prefix,
                "expire": self.expire,
                "show_spectate_link": self.show_spectate_link,
                "channels": [channel.channel_xid for channel in self.channels],
                "teams": [team.name for team in self.teams],
            }
        )


class Team(Base):
    __tablename__ = "teams"
    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    guild_xid = Column(
        BigInteger,
        ForeignKey("servers.guild_xid", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    name = Column(String(50), nullable=False)
    server = relationship("Server", back_populates="teams")

    @classmethod
    def points(cls, session: Session, guild_xid: int) -> dict:
        rows = (
            session.query(Team.name, func.sum(UserPoints.points))
            .select_from(User)
            .join(UserServerSettings)
            .join(UserPoints)
            .join(Team)
            .filter(Team.guild_xid == guild_xid)
            .group_by(Team.name)
            .all()
        )
        server = session.query(Server).filter(Server.guild_xid == guild_xid).one_or_none()
        assert server
        results = {team.name: 0 for team in server.teams}
        for row in rows:
            results[row[0]] = row[1]
        return results


class Channel(Base):
    __tablename__ = "channels"
    channel_xid = Column(BigInteger, primary_key=True, nullable=False)
    guild_xid = Column(
        BigInteger,
        ForeignKey("servers.guild_xid", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    server = relationship("Server", back_populates="channels")


class AutoVerifyChannel(Base):
    __tablename__ = "auto_verify_channels"
    channel_xid = Column(BigInteger, primary_key=True, nullable=False)
    guild_xid = Column(
        BigInteger,
        ForeignKey("servers.guild_xid", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    server = relationship("Server", back_populates="auto_verify_channels")


class UnverifiedOnlyChannel(Base):
    __tablename__ = "unverified_only_channels"
    channel_xid = Column(BigInteger, primary_key=True, nullable=False)
    guild_xid = Column(
        BigInteger,
        ForeignKey("servers.guild_xid", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    server = relationship("Server", back_populates="unverified_only_channels")


class ChannelSettings(Base):
    __tablename__ = "channel_settings"
    channel_xid = Column(BigInteger, primary_key=True, nullable=False)
    guild_xid = Column(
        BigInteger,
        ForeignKey("servers.guild_xid", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    cached_name = Column(String(50))
    created_at = Column(DateTime, nullable=False, default=datetime.utcnow)
    updated_at = Column(DateTime, nullable=False, default=datetime.utcnow)
    default_size = Column(Integer, nullable=True)
    require_verification = Column(Boolean, nullable=False, server_default=false())
    cmotd = Column(String(255))
    verify_message = Column(String(255))
    tags_enabled = Column(Boolean, nullable=True)  # overrides server setting
    queue_time_enabled = Column(Boolean, nullable=True)  # overrides server setting
    server = relationship("Server", back_populates="channel_settings")


games_tags = Table(
    "games_tags",
    Base.metadata,
    Column(
        "game_id", Integer, ForeignKey("games.id", ondelete="CASCADE"), primary_key=True
    ),
    Column(
        "tag_id", Integer, ForeignKey("tags.id", ondelete="CASCADE"), primary_key=True
    ),
)


class Event(Base):
    __tablename__ = "events"
    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    games = relationship("Game", back_populates="event", uselist=True)

    @property
    def started(self) -> bool:
        return any(game.status == "started" for game in self.games)

    def __repr__(self) -> str:
        return json.dumps({"id": self.id})


users_blocks = Table(
    "users_blocks",
    Base.metadata,
    Column(
        "user_xid",
        BigInteger,
        ForeignKey("users.xid", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
        index=True,
    ),
    Column(
        "blocked_user_xid",
        BigInteger,
        ForeignKey("users.xid", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
        index=True,
    ),
    UniqueConstraint("user_xid", "blocked_user_xid", name="uix_1"),
)


class User(Base):
    __tablename__ = "users"
    xid = Column(BigInteger, primary_key=True, nullable=False)
    created_at = Column(DateTime, nullable=False, default=datetime.utcnow)
    updated_at = Column(DateTime, nullable=False, default=datetime.utcnow)
    game_id = Column(
        Integer, ForeignKey("games.id", ondelete="SET NULL"), nullable=True, index=True
    )
    cached_name = Column(String(50))
    power = Column(Integer, nullable=True)
    banned = Column(Boolean, nullable=False, server_default=false())
    game = relationship("Game", back_populates="users")

    blocks = relationship(
        "User",
        secondary=users_blocks,
        primaryjoin=xid == users_blocks.c.user_xid,
        secondaryjoin=xid == users_blocks.c.blocked_user_xid,
        backref="blocked_by",
    )

    def blocked(self, game: Game) -> bool:
        session = Session.object_session(self)
        other_user_xids = [u.xid for u in game.users]
        query = session.query(users_blocks).filter(
            or_(
                and_(
                    users_blocks.c.user_xid == self.xid,
                    users_blocks.c.blocked_user_xid.in_(other_user_xids),
                ),
                and_(
                    users_blocks.c.blocked_user_xid == self.xid,
                    users_blocks.c.user_xid.in_(other_user_xids),
                ),
            )
        )
        return cast(bool, session.query(query.exists()).scalar())

    @classmethod
    def recent_metrics(cls, session: Session) -> dict:
        data = [
            row[1]
            for row in (
                session.query(
                    func.date(User.created_at).label("day"),
                    func.count(User.xid),
                )
                .filter(
                    User.created_at >= datetime.utcnow() - timedelta(days=5),
                )
                .group_by("day")
                .order_by(desc("day"))
                .all()
            )
        ]
        return {f"users_{i}": count for i, count in enumerate(data)}

    @property
    def waiting(self) -> bool:
        if self.game and self.game.status in ["pending", "ready"]:
            return True
        return False

    def points(self, guild_xid) -> int:
        session = Session.object_session(self)
        filters = and_(UserPoints.user_xid == self.xid, UserPoints.guild_xid == guild_xid)
        results = session.query(func.sum(UserPoints.points)).filter(filters).scalar()
        return results or 0

    def to_json(self) -> dict:
        return {
            "cached_name": self.cached_name,
            "created_at": str(self.created_at),
            "game_id": self.game_id,
            "power": self.power,
            "xid": self.xid,
        }


class UserServerSettings(Base):
    __tablename__ = "user_server_settings"
    user_xid = Column(
        BigInteger,
        ForeignKey("users.xid", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    guild_xid = Column(
        BigInteger,
        ForeignKey("servers.guild_xid", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    team_id = Column(Integer, ForeignKey("teams.id", ondelete="CASCADE"), nullable=True)
    verified = Column(Boolean, nullable=True, server_default=false())


class WatchedUser(Base):
    __tablename__ = "watched_users"
    user_xid = Column(
        BigInteger,
        ForeignKey("users.xid", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    guild_xid = Column(
        BigInteger,
        ForeignKey("servers.guild_xid", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    note = Column(String(255))


class UserPoints(Base):
    __tablename__ = "user_points"
    user_xid = Column(
        BigInteger,
        ForeignKey("users.xid", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    guild_xid = Column(
        BigInteger,
        ForeignKey("servers.guild_xid", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    game_id = Column(
        Integer,
        ForeignKey("games.id", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    points = Column(Integer, nullable=False)


class Tag(Base):
    __tablename__ = "tags"
    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    name = Column(String(50), nullable=False)
    games = relationship(
        "Game", secondary=games_tags, back_populates="tags", uselist=True
    )

    @classmethod
    def create_many(cls, session: Session, tag_names: List[str]) -> List[Tag]:
        created_at_least_one = False
        tags = []
        for tag_name in tag_names:
            tag = session.query(Tag).filter_by(name=tag_name).one_or_none()
            if not tag:
                created_at_least_one = True
                tag = Tag(name=tag_name)
                session.add(tag)
            tags.append(tag)
        if created_at_least_one:
            session.commit()
        return tags


class Play(Base):
    __tablename__ = "plays"
    user_xid = Column(
        BigInteger,
        ForeignKey("users.xid", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    game_id = Column(
        Integer,
        ForeignKey("games.id", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )


class Game(Base):
    __tablename__ = "games"
    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    created_at = Column(DateTime, nullable=False, default=datetime.utcnow)
    updated_at = Column(DateTime, default=datetime.utcnow, index=True)
    expires_at = Column(DateTime)
    size = Column(Integer, nullable=False, index=True)
    guild_xid = Column(
        BigInteger,
        ForeignKey("servers.guild_xid", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    channel_xid = Column(BigInteger, index=True)
    url = Column(String(255))
    status = Column(
        String(30), nullable=False, server_default=text("'pending'"), index=True
    )
    message = Column(String(255))
    note = Column(String(255))
    event_id = Column(
        Integer, ForeignKey("events.id", ondelete="SET NULL"), nullable=True, index=True
    )
    message_xid = Column(BigInteger)
    system = Column(
        String(30), nullable=False, server_default=text("'spelltable'"), index=True
    )
    game_power = Column(Float, nullable=True)
    voice_channel_xid = Column(BigInteger, nullable=True, index=True)
    voice_channel_invite = Column(String(255))
    users = relationship("User", back_populates="game", uselist=True)
    tags = relationship("Tag", secondary=games_tags, back_populates="games", uselist=True)
    server = relationship("Server", back_populates="games")
    event = relationship("Event", back_populates="games")
    reports = relationship("Report", back_populates="game", uselist=True)

    @classmethod
    def games_per_day(cls, session: Session) -> Iterable[Tuple[datetime, int]]:
        return cast(
            Iterable[Tuple[datetime, int]],
            session.query(
                func.date(Game.created_at).label("day"),
                func.count(Game.id),
            )
            .filter(
                and_(
                    Game.status == "started",
                    Game.created_at >= datetime.utcnow() - timedelta(days=5),
                )
            )
            .group_by("day")
            .order_by(desc("day"))
            .all(),
        )

    @classmethod
    def games_per_day_per_channel(
        cls, session: Session, guild_xid: int
    ) -> Iterable[Tuple[datetime, int, int]]:
        filters = [
            Game.status == "started",
            Game.guild_xid == guild_xid,
        ]
        return cast(
            Iterable[Tuple[datetime, int, int]],
            session.query(
                func.date(Game.created_at).label("day"),
                Game.channel_xid,
                func.count(Game.id),
            )
            .filter(and_(*filters))
            .group_by("day", Game.channel_xid)
            .order_by(desc("day"), Game.channel_xid)
            .all(),
        )

    @classmethod
    def recent_metrics(cls, session: Session) -> dict:
        data = [row[1] for row in cls.games_per_day(session)]
        return {f"games_{i}": count for i, count in enumerate(data)}

    @property
    def power(self) -> Optional[float]:
        if not self.server.power_enabled:
            return None
        values: List[int] = [user.power for user in self.users if user.power]
        if values:
            return sum(values) / len(values)
        return None

    @classmethod
    def find_existing(
        cls,
        *,
        session: Session,
        server: Server,
        channel_xid: int,
        size: int,
        seats: int,
        tags: List[Tag],
        system: str,
        power: Optional[int],
    ) -> Optional[Game]:
        guild_xid = server.guild_xid
        required_tag_ids = set(tag.id for tag in tags)
        select_filters = [
            Game.status == "pending",
            Game.guild_xid == guild_xid,
            Game.size == size,
            Game.channel_xid == channel_xid,
            Game.system == system,
            Game.message_xid.isnot(None),
        ]
        having_filters = [
            func.count(distinct(games_tags.c.tag_id)) == len(required_tag_ids),
            func.count(distinct(User.xid)) <= size - seats,
        ]
        if server.power_enabled:
            if power:
                having_filters.append(between(func.avg(User.power), power - 1, power + 1))
                if power >= 7:
                    having_filters.append(func.avg(User.power) >= 7)
                else:
                    having_filters.append(func.avg(User.power) < 7)
            else:
                select_filters.append(User.power.is_(None))
        inner = (
            session.query(Game.id)
            .join(User, isouter=True)
            .join(games_tags, isouter=True)
            .filter(and_(*select_filters))
            .group_by(Game.id)
            .having(and_(*having_filters))
        )
        tag_filters = []
        for tid in required_tag_ids:
            tag_filters.append(games_tags.c.tag_id == tid)
        outer = (
            session.query(Game)
            .join(games_tags, isouter=True)
            .filter(and_(Game.id.in_(inner), or_(*tag_filters)))
            .group_by(Game.id)
            .having(having_filters[0])
            .order_by(asc(Game.updated_at))
        )
        game: Optional[Game] = outer.first()
        return game

    @classmethod
    def expired(cls, session: Session) -> List[Game]:
        return cast(
            List[Game],
            session.query(Game)
            .filter(
                and_(
                    datetime.utcnow() >= Game.expires_at,
                    Game.status == "pending",
                )
            )
            .all(),
        )

    def is_expired(self) -> bool:
        if (
            self.status == "pending"
            and self.expires_at
            and datetime.utcnow() >= self.expires_at
        ):
            return True
        return False

    @classmethod
    def voiced(cls, session: Session) -> List[Game]:
        return cast(
            List[Game],
            session.query(Game)
            .filter(
                and_(
                    Game.voice_channel_xid.isnot(None),
                    datetime.utcnow() - timedelta(minutes=10) >= Game.updated_at,
                )
            )
            .all(),
        )

    @classmethod
    def average_wait_times(cls, session: Session):
        if "sqlite" in session.bind.dialect.name:  # pragma: no cover
            avg = func.avg(
                (func.julianday(Game.updated_at) - func.julianday(Game.created_at))
                * 1440.0
            )
        else:  # pragma: no cover
            avg = (
                sql_cast(
                    func.avg(
                        text("EXTRACT(EPOCH FROM(games.updated_at - games.created_at))")
                    ),
                    Numeric,
                )
                / 60.0
            )
        return (
            session.query(
                Game.guild_xid,
                Game.channel_xid,
                avg,
            )
            .filter(
                and_(
                    Game.created_at >= datetime.utcnow() - timedelta(hours=1),
                    Game.status == "started",
                    Game.channel_xid.isnot(None),
                )
            )
            .group_by(Game.guild_xid, Game.channel_xid)
            .all()
        )

    def __repr__(self) -> str:
        return json.dumps(self.to_json())

    def to_json(self) -> dict:
        return {
            "id": self.id,
            "created_at": str(self.created_at),
            "updated_at": str(self.updated_at),
            "expires_at": str(self.expires_at),
            "size": self.size,
            "guild_xid": self.guild_xid,
            "channel_xid": self.channel_xid,
            "url": self.url,
            "status": self.status,
            "system": self.system,
            "message": self.message,
            "note": self.note,
            "message_xid": self.message_xid,
            "voice_channel_xid": self.voice_channel_xid,
            "voice_channel_invite": self.voice_channel_invite,
            "tags": [tag.name for tag in self.tags],
            "event_id": self.event.id if self.event else None,
            "power": self.power,
        }

    def to_embed(self, dm: bool = False, wait: Optional[float] = None) -> discord.Embed:
        prefix = self.server.prefix
        show_motd = (
            True
            if self.server.motd == "both"
            else True
            if dm and self.server.motd == "private"
            else True
            if not dm and self.server.motd == "public"
            else False
        )
        show_link = True if dm else self.server.links == "public"
        if self.status == "pending":
            remaining = int(self.size) - len(cast(List[User], self.users))
            plural = "s" if remaining > 1 else ""
            title = f"**Waiting for {remaining} more player{plural} to join...**"
        else:
            title = self.message if self.message else "**Your game is ready!**"
        embed = discord.Embed(title=title)
        embed.set_thumbnail(url=THUMB_URL)

        description = ""
        if self.voice_channel_invite:
            if show_link:
                description = (
                    f"Join your voice chat now: {self.voice_channel_invite} (invite"
                    f" expires in {int(VOICE_INVITE_EXPIRE_TIME_S / 60)} minutes)\n\n"
                )
        if self.status == "pending":
            description += (
                f"To **join this game**, react with {EMOJI_JOIN_GAME}\n"
                f"If you need to drop, react with {EMOJI_DROP_GAME}\n\n"
                "_A SpellTable link will be created when all players have joined._\n\n"
                f"Looking for more players to join you? Just run `{prefix}lfg` again.\n"
            )
        elif self.system == "spelltable":
            if show_link:
                if self.url:
                    description += (
                        "Click the link below to join your SpellTable game."
                        f"\n<{self.url}>"
                    )
                    if self.server.show_spectate_link:
                        description += (
                            "\nOr spectate on the game with the following link."
                            f"\n<{self.url}?spectate>"
                        )
                else:
                    description += (
                        "Sorry but SpellBot was unable to create a SpellTable link"
                        " for this game. Please go to"
                        " [spelltable.com](https://www.spelltable.com/) to create one.\n"
                    )
            else:
                description += (
                    "Please check your Direct Messages for your SpellTable link."
                )
        elif self.system == "mtgo":
            description += (
                "Please exchange MTGO contact information and head over there to play!"
            )
        else:  # self.system == "arena"
            description += (
                "Please exchange Arena contact information and head over there to play!"
            )
        if self.channel_xid:
            channel_settings = self.server.channel_settings_for(self.channel_xid)
            if channel_settings and channel_settings.cmotd:
                description += f"\n{channel_settings.cmotd}"
        if self.server.smotd and show_motd:
            description += f"\n{self.server.smotd}"
        embed.description = description

        if self.voice_channel_xid:
            embed.add_field(name="Voice Channel", value=f"<#{self.voice_channel_xid}>")
        tag_names = None
        if self.tags and len(cast(List[Tag], self.tags)) >= 1:
            sorted_tag_names: List[str] = sorted([tag.name for tag in self.tags])
            tag_names = ", ".join(sorted_tag_names)
            embed.add_field(name="Tags", value=tag_names)
        power = self.power
        if power and self.server.power_enabled:
            embed.add_field(name="Average Power Level", value=f"{power:.1f}")
        if wait:
            delta = timedelta(minutes=wait)
            mins = humanize.precisedelta(delta, minimum_unit="seconds", format="%0.0f")
            embed.add_field(name="Average Queue Time", value=mins)
        if self.users:
            players = ", ".join(
                sorted(
                    [
                        f"<@{user.xid}>"  # Support <@!USERID> for server nick?
                        + (
                            f" (Power: {user.power})"
                            if (user.power and self.server.power_enabled)
                            else ""
                        )
                        for user in self.users
                    ]
                )
            )
            embed.add_field(name="Players", value=players, inline=False)

        if self.note:
            embed.add_field(name="Note", value=self.note, inline=False)

        embed.set_footer(text=f"SpellBot Reference #SB{self.id}")
        embed.color = discord.Color(0x5A3EFD)
        return embed


class Report(Base):
    __tablename__ = "reports"
    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    game_id = Column(
        Integer, ForeignKey("games.id", ondelete="CASCADE"), nullable=False, index=True
    )
    report = Column(String(255), nullable=False)
    created_at = Column(DateTime, nullable=False, default=datetime.utcnow)
    game = relationship("Game", back_populates="reports")


def create_all(connection: Connection, db_url: str) -> None:
    config = alembic.config.Config(str(ALEMBIC_INI))
    config.set_main_option("script_location", str(VERSIONS_DIR))
    config.set_main_option("sqlalchemy.url", db_url)
    cast(Any, config.attributes)["connection"] = connection
    alembic.command.upgrade(config, "head")


def reverse_all(connection: Connection, db_url: str) -> None:
    config = alembic.config.Config(str(ALEMBIC_INI))
    config.set_main_option("script_location", str(VERSIONS_DIR))
    config.set_main_option("sqlalchemy.url", db_url)
    cast(Any, config.attributes)["connection"] = connection
    alembic.command.downgrade(config, "base")


class Metric(Base):
    __tablename__ = "metrics"
    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    created_at = Column(DateTime, nullable=False, default=datetime.utcnow)
    kind = Column(String(50), index=True)
    guild_xid = Column(BigInteger, index=True)
    channel_xid = Column(BigInteger, index=True)
    user_xid = Column(BigInteger, index=True)


class Data:
    """Persistent and in-memory store for user data."""

    def __init__(self, db_url: str):
        self.db_url = db_url
        self.engine = create_engine(db_url, echo=False)
        self.conn = self.engine.connect()
        create_all(self.conn, db_url)
        self.Session = sessionmaker(bind=self.engine)
        self.metadata = Base.metadata



"""Add user blocks

Revision ID: bef35ca63436
Revises: 84903dda8911
Create Date: 2021-01-22 14:41:08.839592

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "bef35ca63436"
down_revision = "84903dda8911"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "users_blocks",
        sa.Column("user_xid", sa.BigInteger(), nullable=False),
        sa.Column("blocked_user_xid", sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(["blocked_user_xid"], ["users.xid"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["user_xid"], ["users.xid"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("user_xid", "blocked_user_xid"),
        sa.UniqueConstraint("user_xid", "blocked_user_xid", name="uix_1"),
    )
    op.create_index(
        op.f("ix_users_blocks_blocked_user_xid"),
        "users_blocks",
        ["blocked_user_xid"],
        unique=False,
    )
    op.create_index(
        op.f("ix_users_blocks_user_xid"), "users_blocks", ["user_xid"], unique=False
    )


def downgrade():
    op.drop_index(op.f("ix_users_blocks_user_xid"), table_name="users_blocks")
    op.drop_index(op.f("ix_users_blocks_blocked_user_xid"), table_name="users_blocks")
    op.drop_table("users_blocks")



"""Refactor to replace queue logic with lfg

Revision ID: d5208ea8d47f
Revises: 6a1ea19d138f
Create Date: 2020-07-15 12:51:40.082517

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d5208ea8d47f"
down_revision = "6a1ea19d138f"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("friendly")
        b.drop_column("scope")


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(
            sa.Column(
                "scope",
                sa.VARCHAR(length=10),
                nullable=False,
                server_default=sa.text("'server'"),
            )
        )
        b.add_column(sa.Column("friendly", sa.BOOLEAN(), nullable=True))



from datetime import datetime

ADMIN_ROLE = "SpellBot Admin"
CREATE_ENDPOINT = "https://us-central1-magic-night-30324.cloudfunctions.net/createGame"
THUMB_URL = (
    "https://raw.githubusercontent.com/lexicalunit/spellbot/master/spellbot.png"
    f"?{datetime.today().strftime('%Y-%m-%d')}"  # workaround over-eager caching
)
ICO_URL = (
    "https://raw.githubusercontent.com/lexicalunit/spellbot/master/spellbot-sm.png"
    f"?{datetime.today().strftime('%Y-%m-%d')}"  # workaround over-eager caching
)
DEFAULT_GAME_SIZE = 4
INVITE_LINK = (
    "https://discordapp.com/api/oauth2/authorize"
    "?client_id=725510263251402832&permissions=93265&scope=bot"
)
VOTE_LINK = "https://top.gg/bot/725510263251402832/vote"

EMOJI_DROP_GAME = "🚫"
EMOJI_FAIL = "❌"
EMOJI_JOIN_GAME = "✋"
EMOJI_OK = "✅"

VOICE_CATEGORY_PREFIX = "SpellBot Voice Channels"

VOICE_INVITE_EXPIRE_TIME_S = 600  # ten minutes, make this configurable?
CLEAN_S = 7  # seconds before temporary messages by the bot get cleaned up, configurable?



"""Adds user power level

Revision ID: ba6af6c62640
Revises: 7a2698d54c8a
Create Date: 2020-08-14 14:49:21.107946

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ba6af6c62640"
down_revision = "7a2698d54c8a"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("users") as b:
        b.add_column(sa.Column("power", sa.Integer(), nullable=True))


def downgrade():
    with op.batch_alter_table("users") as b:
        b.drop_column("power")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['spellapi', 'spellbot', 'spellbot.versions', 'spellbot.versions.versions']

package_data = \
{'': ['*'], 'spellbot': ['assets/*']}

install_requires = \
['aiofiles==0.6.0',
 'aiohttp==3.6.3',
 'aioredis==1.3.1',
 'alembic==1.5.8',
 'click==7.1.2',
 'coloredlogs==15.0',
 'discord.py==1.7.1',
 'dunamai==1.5.4',
 'expiringdict==1.2.1',
 'fastapi-cache2==0.1.3.4',
 'fastapi>=0.62,<0.64',
 'httpx==0.17.1',
 'humanize==3.3.0',
 'hupper==1.10.2',
 'itsdangerous==1.1.0',
 'psycopg2-binary==2.8.6',
 'python-dateutil==2.8.1',
 'python-dotenv==0.16.0',
 'python-jose[cryptography]==3.2.0',
 'pytz==2021.1',
 'pyyaml==5.4.1',
 'redis==3.5.3',
 'requests==2.25.1',
 'sqlalchemy==1.4.4',
 'unidecode==1.2.0',
 'uvicorn==0.13.4']

entry_points = \
{'console_scripts': ['spellapi = spellapi:main', 'spellbot = spellbot:main']}

setup_kwargs = {
    'name': 'spellbot',
    'version': '5.27.2',
    'description': 'The Discord bot for SpellTable',
    'long_description': '<img align="right" width="200" src="https://raw.githubusercontent.com/lexicalunit/spellbot/master/spellbot.png" />\n\n# SpellBot\n\n[![build][build-badge]][build]\n[![top][top-badge]][top]\n[![uptime][uptime-badge]][uptime]\n[![metrics][metrics-badge]][metrics]\n[![pypi][pypi-badge]][pypi]\n[![codecov][codecov-badge]][codecov]\n[![CodeFactor][factor-badge]][factor]\n[![CodeQL][codeql-badge]][security]\n[![python][python-badge]][python]\n[![black][black-badge]][black]\n[![mit][mit-badge]][mit]\n[![patreon][patreon-button]][patreon]\n[![follow][follow-badge]][follow]\n\nThe Discord bot for [SpellTable][spelltable].\n\n[![add-bot][add-bot-button]][add-bot]\n\n## 🤖 Using SpellBot\n\nOnce you\'ve connected the bot to your server, you can interact with it over\nDiscord via the following commands in any of the authorized channels. **Keep in\nmind that sometimes SpellBot will respond to you via Direct Message to avoid\nbeing too spammy in text channels.**\n\n- `!spellbot help`: Provides detailed help about all of the following commands.\n- `!about`: Get information about SpellBot and its creators.\n\n> **Note:** To use the\n> [commands for event runners](#%EF%B8%8F-commands-for-event-runners)\n> and [commands for admins](#-commands-for-admins), you will need to\n> [create a role on your server][create-role] called `SpellBot Admin`\n> (capitalization matters). Only users with that role will be able to use those\n> commands.\n\n### ✋ Commands for Players\n\nJust looking to play some games of Magic? These commands are for you!\n\n- `!lfg`: Find or start up a game of Magic: The Gathering!\n- `!leave`: Leave any games that you\'ve signed up for.\n- `!power`: Set your current power level.\n- `!report`: Report the results of the game you just played.\n- `!team`: Join one of the teams available on your server.\n- `!points`: Find out how many points you currently have.\n- `!block`: Block other users from joining your games.\n- `!unblock`: Unblock previously blocked users.\n\nWhen you run the `!lfg` command, SpellBot will post a message for sign ups.\n\n![lfg][lfg]\n\nOther users can react to it with the ✋ emoji to be added to the game. When the\ngame is ready, SpellBot will update the message with your SpellTable details.\n\n![ready][ready]\n\nUsers can also use the 🚫 emoji reaction to leave the game.\n\n### 🎟️ Commands for Event Runners\n\nThese commands are intended to be run by SpellBot Admins and help facilitate\nonline events.\n\n- `!game`: Directly create games for the mentioned users.\n- `!event`: Create a bunch of games all at once based on some uploaded data.\n- `!begin`: Start an event that you previously created with `!event`.\n- `!export`: Export historical game data for your server.\n\n### 👑 Commands for Admins\n\nThese commands will help you configure SpellBot for your server.\n\n- `!spellbot`: This command allows admins to configure SpellBot for their\n               server. It supports the following subcommands:\n  - `config`: Just show the current configuration for this server.\n  - `channels`: Set the channels SpellBot is allowed to operate within.\n  - `prefix`: Set the command prefix for SpellBot in text channels.\n  - `links`: Set the privacy level for generated SpellTable links.\n  - `spectate`: Add a spectator link to the posts SpellBot makes.\n  - `expire`: Set how many minutes before games are expired due to inactivity.\n  - `teams`: Sets the teams available on this server.\n  - `power`: Turns the power command on or off for this server.\n  - `voice`: When on, SpellBot will automatically create voice channels.\n  - `tags`: Turn on or off the ability to use tags. Optionally mention specific channels.\n  - `queue-time`: Turn on or off average queue time details. Optionally mention specific channels.\n  - `smotd`: Set the server message of the day.\n  - `cmotd`: Set the message of the day for the channel where you run it.\n  - `motd`: Set the privacy level for messages of the day.\n  - `size`: Sets the default game size for a specific channel.\n  - `toggle-verify`: Toggles requirement of verification for a specific channel.\n  - `auto-verify`: Set the channels that will trigger user auto verification.\n  - `unverified-only`: Set the channels that are only for unverified users.\n  - `verify-message`: Set the verification message for a specific channel.\n  - `voice-category`: Set category for voice channels created by !game.\n  - `stats`: Gets some statistics about SpellBot usage on your server.\n  - `help`: Get detailed usage help for SpellBot.\n- `!verify`: Allows moderators to verify a user on their server.\n- `!unverify`: Un-verifies a user for this server.\n- `!watch`: Allows moderators to watch a user on their server.\n- `!unwatch`: Un-watches a user for this server.\n\n### 🛋️ Ergonomics\n\nSpellBot will always try and assume useful defaults or try to do the right thing\nwhen you give it a command. For example if you use the tag <code>~modern</code>\nor other format names when creating a game, it will automatically assume the\ncorrect number of players for you. Hopefully these features are intuitive and\nhelpful 🤞 — and if not, [please report bugs and request features][issues]\nto your heart\'s content.\n\n### 🎤 Feedback\n\nThoughts and suggestions? Come join us on the\n[SpellTable Discord server][discord-invite]! Please also feel free\nto [directly report any bugs][issues] that you encounter.\n\n## 🙌 Support Me\n\nI\'m keeping SpellBot running using my own money but if you like the bot and want\nto help me out, please consider [becoming a patron][patreon].\n\n## ❤️ Contributing\n\nIf you\'d like to become a part of the SpellBot development community please\nfirst know that we have a documented [code of conduct](CODE_OF_CONDUCT.md) and\nthen see our [documentation on how to contribute](CONTRIBUTING.md) for details\non how to get started.\n\n## 🔧 Running SpellBot Yourself\n\nFirst install `spellbot` using [`pip`](https://pip.pypa.io/en/stable/):\n\n```shell\npip install spellbot\n```\n\nProvide your Discord bot token with the environment variable `SPELLBOT_TOKEN`.\nAs well as your SpellTable API authorization token via `SPELLTABLE_AUTH`.\n\nYou can get [your bot token from Discord][discord-bot-docs]. Your bot will\nneed the following permissions enabled:\n\n- Manage Channels\n- Create Instant Invite\n- View Channels\n- Send Messages\n- Manage Messages\n- Embed Links\n- Read Message History\n- Add Reactions\n\nAs for the SpellTable API authorization token, you\'ll have to talk to the\nSpellTable developers. Come join us on the\n[SpellTable Discord server][spelltable-discord].\n\nBy default SpellBot will use sqlite3 as its database. You can however choose to\nuse another database by providing a [SQLAlchemy Connection URL][db-url]. This\ncan be done via the `--database-url` command line option or the environment\nvariable `SPELLBOT_DB_URL`. Note that, at the time of this writing, SpellBot is\nonly tested against sqlite3 and PostgreSQL.\n\nMore usage help can be found by running `spellbot --help`.\n\n## 🐳 Docker Support\n\nYou can also run SpellBot via docker. Our image is published to\n[lexicalunit/spellbot][docker-hub]. See [our documentation on Docker Support](DOCKER.md) for help\nwith installing and using it.\n\n---\n\n[MIT][mit] © [amy@lexicalunit][lexicalunit] et [al][contributors]\n\n[add-bot-button]:     https://user-images.githubusercontent.com/1903876/88951823-5d6c9a00-d24b-11ea-8523-d256ccbf4a3c.png\n[add-bot]:            https://discordapp.com/api/oauth2/authorize?client_id=725510263251402832&permissions=93265&scope=bot\n[black-badge]:        https://img.shields.io/badge/code%20style-black-000000.svg\n[black]:              https://github.com/psf/black\n[build-badge]:        https://github.com/lexicalunit/spellbot/workflows/build/badge.svg\n[build]:              https://github.com/lexicalunit/spellbot/actions\n[codecov-badge]:      https://codecov.io/gh/lexicalunit/spellbot/branch/master/graph/badge.svg\n[codecov]:            https://codecov.io/gh/lexicalunit/spellbot\n[codeql-badge]:       https://github.com/lexicalunit/spellbot/workflows/CodeQL/badge.svg\n[contributors]:       https://github.com/lexicalunit/spellbot/graphs/contributors\n[create-role]:        https://support.discord.com/hc/en-us/articles/206029707-How-do-I-set-up-Permissions-\n[db-url]:             https://docs.sqlalchemy.org/en/latest/core/engines.html\n[discord-bot-docs]:   https://discord.com/developers/docs/topics/oauth2#bots\n[discord-invite]:     https://discord.gg/zXzgqMN\n[docker-hub]:         https://hub.docker.com/r/lexicalunit/spellbot\n[factor-badge]:       https://www.codefactor.io/repository/github/lexicalunit/spellbot/badge\n[factor]:             https://www.codefactor.io/repository/github/lexicalunit/spellbot\n[follow-badge]:       https://img.shields.io/twitter/follow/SpellBotIO?style=social\n[follow]:             https://twitter.com/intent/follow?screen_name=SpellBotIO\n[issues]:             https://github.com/lexicalunit/spellbot/issues\n[patreon]:            https://www.patreon.com/lexicalunit\n[patreon-button]:     https://img.shields.io/endpoint.svg?url=https%3A%2F%2Fshieldsio-patreon.vercel.app%2Fapi%3Fusername%3Dlexicalunit%26type%3Dpatrons88951826-5e053080-d24b-11ea-9a81-f1b5431a5d4b.png\n[lexicalunit]:        http://github.com/lexicalunit\n[lfg]:                https://user-images.githubusercontent.com/1903876/91242259-cedd2280-e6fb-11ea-8d30-e7127b6f96e9.png\n[metrics-badge]:      https://img.shields.io/badge/metrics-grafana-orange.svg\n[metrics]:            https://lexicalunit.grafana.net/d/4TSUCbcMz/spellbot?orgId=1\n[mit-badge]:          https://img.shields.io/badge/License-MIT-yellow.svg\n[mit]:                https://opensource.org/licenses/MIT\n[pypi-badge]:         https://img.shields.io/pypi/v/spellbot\n[pypi]:               https://pypi.org/project/spellbot/\n[python-badge]:       https://img.shields.io/badge/python-3.8+-blue.svg\n[python]:             https://www.python.org/\n[ready]:              https://user-images.githubusercontent.com/1903876/91242257-cdabf580-e6fb-11ea-86ad-8f1aaf6d34dc.png\n[security]:           https://github.com/lexicalunit/spellbot/security\n[spelltable-discord]: https://discord.gg/zXzgqMN\n[spelltable]:         https://www.spelltable.com/\n[top-badge]:          https://top.gg/api/widget/status/725510263251402832.svg?noavatar=true\n[top]:                https://top.gg/bot/725510263251402832\n[uptime-badge]:       https://img.shields.io/uptimerobot/ratio/m785764282-c51c742e56a87d802968efcc\n[uptime]:             https://uptimerobot.com/dashboard#785764282\n',
    'author': 'Amy',
    'author_email': 'amy@lexicalunit.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://spellbot.io/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



"""Add verified column to user server settings

Revision ID: aa4e7f42cd91
Revises: da6dc39b4882
Create Date: 2020-11-19 15:04:17.953982

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "aa4e7f42cd91"
down_revision = "da6dc39b4882"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("user_server_settings") as b:
        b.add_column(
            sa.Column("verified", sa.Boolean(), server_default=sa.false(), nullable=True),
        )


def downgrade():
    with op.batch_alter_table("user_server_settings") as b:
        b.drop_column("verified")



"""Rebuild

Revision ID: dc915aa5af34
Revises:
Create Date: 2020-07-04 19:30:52.888268

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "dc915aa5af34"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "servers",
        sa.Column("guild_xid", sa.BigInteger(), nullable=False),
        sa.Column("prefix", sa.String(length=10), nullable=False),
        sa.Column("scope", sa.String(length=10), nullable=False),
        sa.Column("expire", sa.Integer(), server_default=sa.text("30"), nullable=False),
        sa.PrimaryKeyConstraint("guild_xid"),
    )
    op.create_table(
        "tags",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("name", sa.String(length=50), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "wait_times",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("guild_xid", sa.BigInteger(), nullable=False),
        sa.Column("channel_xid", sa.BigInteger(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("seconds", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "authorized_channels",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("guild_xid", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.ForeignKeyConstraint(["guild_xid"], ["servers.guild_xid"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "games",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("size", sa.Integer(), nullable=False),
        sa.Column("guild_xid", sa.BigInteger(), nullable=False),
        sa.Column("channel_xid", sa.BigInteger(), nullable=True),
        sa.Column("power", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["guild_xid"], ["servers.guild_xid"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "games_tags",
        sa.Column("game_id", sa.Integer(), nullable=True),
        sa.Column("tag_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["game_id"], ["games.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["tag_id"], ["tags.id"], ondelete="CASCADE"),
    )
    op.create_table(
        "users",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("xid", sa.BigInteger(), nullable=False),
        sa.Column("game_id", sa.Integer(), nullable=True),
        sa.Column("queued_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(["game_id"], ["games.id"], ondelete="SET NULL"),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    op.drop_table("users")
    op.drop_table("games_tags")
    op.drop_table("games")
    op.drop_table("authorized_channels")
    op.drop_table("wait_times")
    op.drop_table("tags")
    op.drop_table("servers")



"""Adds voice channels category

Revision ID: e3ab9447b076
Revises: 311f7bbcdf77
Create Date: 2020-10-12 16:06:46.843494

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e3ab9447b076"
down_revision = "311f7bbcdf77"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(sa.Column("voice_category_xid", sa.BigInteger(), nullable=True))


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("voice_category_xid")



"""Adds historical game power tracking

Revision ID: 60f9ffc05b0a
Revises: 95106b5a352d
Create Date: 2020-08-22 08:17:13.008010

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "60f9ffc05b0a"
down_revision = "95106b5a352d"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("games") as b:
        b.add_column(sa.Column("game_power", sa.Float(), nullable=True))


def downgrade():
    with op.batch_alter_table("games") as b:
        b.drop_column("game_power")



"""Adds power command settings

Revision ID: 055832354330
Revises: a711319d9661
Create Date: 2020-09-22 15:12:17.097386

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "055832354330"
down_revision = "a711319d9661"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(
            sa.Column(
                "power_enabled", sa.Boolean(), server_default=sa.true(), nullable=False
            ),
        )


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("power_enabled")



"""Add custom voice category

Revision ID: d7a58c7aaeec
Revises: bef35ca63436
Create Date: 2021-01-28 11:31:02.043758

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d7a58c7aaeec"
down_revision = "bef35ca63436"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(
            sa.Column("voice_category_prefix", sa.String(length=40), nullable=True)
        )


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("voice_category_prefix")



"""Adds voice channel invite url

Revision ID: 8cb01e86be05
Revises: e3ab9447b076
Create Date: 2020-10-13 09:49:54.983214

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8cb01e86be05"
down_revision = "e3ab9447b076"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("games") as b:
        b.add_column(
            sa.Column("voice_channel_invite", sa.String(length=255), nullable=True),
        )


def downgrade():
    with op.batch_alter_table("games") as b:
        b.drop_column("voice_channel_invite")



"""Allow enable/disable average queue times

Revision ID: bf9689a382b3
Revises: 42ab4d026574
Create Date: 2021-02-19 12:41:02.393856

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "bf9689a382b3"
down_revision = "42ab4d026574"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("channel_settings") as b:
        b.add_column(sa.Column("queue_time_enabled", sa.Boolean(), nullable=True))
    with op.batch_alter_table("servers") as b:
        b.add_column(
            sa.Column(
                "queue_time_enabled",
                sa.Boolean(),
                server_default=sa.true(),
                nullable=False,
            ),
        )


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("queue_time_enabled")
    with op.batch_alter_table("channel_settings") as b:
        b.drop_column("queue_time_enabled")



"""Add motd privacy setting

Revision ID: ad589a3d7585
Revises: ba58c2461a59
Create Date: 2020-10-24 20:41:03.351108

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ad589a3d7585"
down_revision = "ba58c2461a59"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(
            sa.Column(
                "motd",
                sa.String(length=10),
                server_default=sa.text("'both'"),
                nullable=False,
            ),
        )


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("motd")



"""Adds user teams

Revision ID: be8afab17cbf
Revises: 95d86da7b3c7
Create Date: 2020-08-19 14:20:57.539270

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "be8afab17cbf"
down_revision = "95d86da7b3c7"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "user_teams",
        sa.Column("user_xid", sa.BigInteger(), nullable=False),
        sa.Column("guild_xid", sa.BigInteger(), nullable=False),
        sa.Column("team_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["guild_xid"], ["servers.guild_xid"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["team_id"], ["teams.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["user_xid"], ["users.xid"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("user_xid", "guild_xid"),
    )


def downgrade():
    op.drop_table("user_teams")



"""Add friendly column to servers table

Revision ID: 64d847efbe4a
Revises: db42410953c4
Create Date: 2020-07-11 16:11:04.621100

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "64d847efbe4a"
down_revision = "db42410953c4"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("servers") as b:
        b.add_column(sa.Column("friendly", sa.Boolean(), nullable=True))


def downgrade():
    with op.batch_alter_table("servers") as b:
        b.drop_column("friendly")
